

#if LIBBLADERF_PRESENT == 1
#include <libbladeRF.h>
#include "globdef.h"
#include "uidef.h"
#include "sdrdef.h"

P_BLADERF bladerf;

char *bladerf_parm_text[MAX_BLADERF_PARM]={"Rx sampling speed",
                                     "Serno1",
                                     "Serno2",
                                     "Serno3",
                                     "Serno4",
                                     "Serno5",
                                     "Serno6",
                                     "Serno7",
                                     "Serno8",
                                     "Freq adjust",
                                     "Sampling",
                                     "Check"};

char *bladerf_parfil_name="par_bladerf";


 /* 
 
// Set the range of VGA1, VGA1GAINT[7:0] 
 + this->vga1_range = osmosdr::gain_range_t( -35, -4, 1 );
 +
 + // Set the range of VGA2, VGA2GAIN[4:0] 
 + this->vga2_range = osmosdr::gain_range_t( 0, 25, 1 );
 +
 + this->setup_device();
 + this->thread = gr::thread::thread(write_task_dispatch, this);
 +}
 +
*/


static inline const char *backend2str(bladerf_backend b)
{
switch (b) 
  {
  case BLADERF_BACKEND_LIBUSB:
  return "libusb";

  case BLADERF_BACKEND_LINUX:
  return "Linux kernel driver";

  default:
  return "Unknown";
  }
}
                                                                    
static inline const char *fpga_size2str(bladerf_fpga_size b)
{
switch (b) 
  {
  case BLADERF_FPGA_40KLE:
  return "40k LE";

  case BLADERF_FPGA_115KLE:
  return "115 kLE";

  default:
  return "Unknown";
  }
}


void bladerf_input(void)
{
}

void init_bladerf(void)
{
int i, err, n_devices, devno, line;
char s[120];
int *ise;
bladerf_fpga_size fpga_size;
bladerf_sampling sampling;
struct bladerf_version version;
struct bladerf_devinfo *devices = NULL;    
struct bladerf *dev;    

ise=(int*)(void*)s;
n_devices = bladerf_get_device_list(&devices);
if (n_devices < 0) 
  {
  if (n_devices == BLADERF_ERR_NODEV) 
    {
    lir_text(5,5,"No bladeRF devices found.");
    } 
  else 
    {
    sprintf(s,"Failed to probe for devices: %s",bladerf_strerror(n_devices));
    lir_text(5,5,s);
    }
  lir_text(5,7,press_any_key);
  await_keyboard();
  return;
  }
line=2;
for(i = 0; i<n_devices; i++) 
  {
  sprintf(s,"%d BladeRF:%s Ser:%s Bus/Addr: %d/%d ", 
     i, backend2str(devices[i].backend), devices[i].serial,
                                devices[i].usb_bus,devices[i].usb_addr);
  lir_text(3,line,s);
  line++;
  }
line++;
lir_text(3, line, "Select device by line number:");
devno=lir_get_integer(32,line,2,0,n_devices-1);
sprintf(s,"%s",devices[devno].serial);
bladerf.sernum1=ise[0];
bladerf.sernum2=ise[1];
bladerf.sernum3=ise[2];
bladerf.sernum4=ise[3];
bladerf.sernum5=ise[4];
bladerf.sernum6=ise[5];
bladerf.sernum7=ise[6];
bladerf.sernum8=ise[7];
err=bladerf_open_with_devinfo(&dev, &devices[devno]);
bladerf_free_device_list(devices);
if(err<0)
  {
#if(OSNUM == OSNUM_WINDOWS)
  lirerr(1370);
#else
  lirerr(1369);
#endif
  return;
  }
i=bladerf_is_fpga_configured(dev);
if(i < 0)
  {
  lirerr(1373);
  goto errexit;
  }
if(i == 0)
  {
  lirerr(1374);
  goto errexit;
  }
version.minor=99;
err=bladerf_fpga_version(dev, &version);
if(err < 0)
  {
  lirerr(1375);
  goto errexit;
  }
err=bladerf_get_fpga_size(dev, &fpga_size);
if(err<0)
  {
  lirerr(1371);
  goto errexit;
  }
sprintf(s,"FPGA size: %s   FPGA version %d.%d.%d (%s),", 
                      fpga_size2str(fpga_size), version.major, 
                         version.minor, version.patch, version.describe);
lir_text(3,line,s);
line++;
err=bladerf_fw_version(dev, &version);
if(err<0)
  {
  lirerr(1372);
  goto errexit;
  }
sprintf(s,"FX3 firmware version %d.%d.%d (%s),", version.major, 
                         version.minor, version.patch, version.describe);
lir_text(3,line,s);
line+=2;
sprintf(s,"Connect ADC directly to input connector? (Y/N)");
lir_text(3,line,s);
getsamp:;
await_processed_keyboard();
if(lir_inkey == 'Y')
  {
  sampling=BLADERF_SAMPLING_EXTERNAL;
  }
else
  {
  if(lir_inkey != 'N')goto getsamp;  
  sampling=BLADERF_SAMPLING_INTERNAL;
  }
bladerf.sampling=(int)sampling;
err=bladerf_set_sampling(dev, sampling);
if(err<0)
  {
  lirerr(1372);
  goto errexit;
  }
  

await_keyboard();




errexit:;
bladerf_close(dev);
/*

getspeed:;
sprintf(s,"Set sampling speed in Hz (%d to %d or %d to %d) =>",
                                    MIN_RTL2832_SAMP_RATE_L,MAX_RTL2832_SAMP_RATE_L,
                                    MIN_RTL2832_SAMP_RATE_H,MAX_RTL2832_SAMP_RATE_H);
lir_text(3,line,s);
rtl2832.sampling_speed=lir_get_integer(4+strlen(s),line,7,
                                MIN_RTL2832_SAMP_RATE_L,MAX_RTL2832_SAMP_RATE_H);
line+=2;
if(rtl2832.sampling_speed > MAX_RTL2832_SAMP_RATE_L &&
   rtl2832.sampling_speed < MIN_RTL2832_SAMP_RATE_H)
  {
  if( rtl2832.sampling_speed-MAX_RTL2832_SAMP_RATE_L >
      MIN_RTL2832_SAMP_RATE_H-rtl2832.sampling_speed)
    {
    rtl2832.sampling_speed=MIN_RTL2832_SAMP_RATE_H;
    }
  else       
    {
    rtl2832.sampling_speed=MAX_RTL2832_SAMP_RATE_L;
    }
  }  
if(kill_all_flag)
  {
  rtlsdr_close(dev_rtlsdr);
  return;
  }
i=rtlsdr_set_sample_rate(dev_rtlsdr, (uint32_t)rtl2832.sampling_speed);
rtl2832.sampling_speed=rtlsdr_get_sample_rate(dev_rtlsdr);
rtlsdr_close(dev_rtlsdr);
if(i < 0)
  {
  lir_text(3,line,"Failed to set sampling rate");
  lir_text(3,line+1,press_any_key);
  await_keyboard();
  clear_lines(line,line+1);
  goto getspeed;
  }

*/
}

int display_bladerf_parm_info(int *line)
{
char s[80];
int errcod;
errcod=read_sdrpar(bladerf_parfil_name, MAX_BLADERF_PARM, 
                            bladerf_parm_text, (int*)((void*)&bladerf));
//if(FBDEV == 0)fflush(NULL);
if(errcod == 0)
  {
  settextcolor(7);
  sprintf(s,"sampling rate = %i, Gain mode = %d, Xtal adjust = %.0f ppb",  
                              rtl2832.sampling_speed, rtl2832.gain_mode,
                              10.*rtl2832.freq_adjust);
  lir_text(24,line[0],s);
  SNDLOG"\n%s",s);
//  line[0]+=1;
  }
  
return (errcod);

}

#endif

